# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

MODEL_CONFIG_ROOT_DIR_PLACEHOLDER = "MODEL_CONFIG_ROOT_DIR"
